package net.knarcraft.questmobspawns.listener;

import me.pikamug.quests.Quests;
import me.pikamug.quests.player.Quester;
import me.pikamug.quests.quests.Quest;
import me.pikamug.quests.quests.components.Stage;
import net.knarcraft.questmobspawns.util.QuestMobHelper;
import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;
import org.bukkit.event.Cancellable;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.EntityBreakDoorEvent;
import org.bukkit.event.entity.EntityChangeBlockEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityExplodeEvent;
import org.bukkit.event.entity.EntityInteractEvent;
import org.bukkit.event.entity.EntityTargetLivingEntityEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * A listener for quest mobs causing damage
 */
public class MobBlockDamageListener implements Listener {

    private final Quests questsAPI;

    public MobBlockDamageListener(@NotNull Quests questsAPI) {
        this.questsAPI = questsAPI;
    }

    @EventHandler(priority = EventPriority.HIGHEST)
    public void onExplode(@NotNull EntityExplodeEvent event) {
        if (QuestMobHelper.isSpawnedMob(event.getEntity())) {
            event.blockList().clear();
            event.setCancelled(false);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockChange(@NotNull EntityChangeBlockEvent event) {
        cancelIfQuestMob(event, event.getEntity());
    }

    @EventHandler
    public void onDoorBreak(@NotNull EntityBreakDoorEvent event) {
        cancelIfQuestMob(event, event.getEntity());
    }

    @EventHandler
    public void onEntityTarget(@NotNull EntityTargetLivingEntityEvent event) {
        Entity entity = event.getEntity();
        if (!QuestMobHelper.isSpawnedMob(entity) || !(event.getTarget() instanceof Player player)) {
            return;
        }

        if (!hasQuest(entity, player)) {
            event.setCancelled(true);
        } else {
            // Cancel the targeting if the player is outside the kill-zone
            int[] killLocationData = QuestMobHelper.getKillZoneLocation(entity);
            Location killLocation = new Location(entity.getWorld(), killLocationData[0], killLocationData[1],
                    killLocationData[2]);
            event.setCancelled(QuestMobHelper.get2dDistance(player.getLocation(), killLocation) > killLocationData[3]);
        }
    }

    @EventHandler
    public void onTrample(@NotNull EntityInteractEvent event) {
        cancelIfQuestMob(event, event.getEntity());
    }

    @EventHandler(priority = EventPriority.HIGHEST)
    public void onEntityDamage(@NotNull EntityDamageByEntityEvent event) {
        if (!QuestMobHelper.isSpawnedMob(event.getDamager())) {
            return;
        }

        Entity damagingEntity = event.getDamager();
        if (event.getEntity() instanceof Player player) {
            event.setCancelled(!hasQuest(damagingEntity, player));
        } else {
            event.setCancelled(true);
        }
    }

    /**
     * Checks whether the given player has the quest the given entity was spawned for
     *
     * @param entity <p>The entity with stored quest data</p>
     * @param player <p>A player</p>
     * @return <p>True if the player has the quest stored in the entity</p>
     */
    private Boolean hasQuest(@NotNull Entity entity, @NotNull Player player) {
        Quester questPlayer = questsAPI.getQuester(player.getUniqueId());
        String questId = QuestMobHelper.getQuestId(entity);
        if (questId == null) {
            return false;
        }
        Quest quest = getQuest(questId);
        if (quest == null) {
            // Remove the entity if the quest it belongs to no longer exists
            entity.remove();
            return false;
        }
        Integer stageId = QuestMobHelper.getStageId(entity);
        if (stageId == null) {
            return false;
        }
        Stage stage = quest.getStage(stageId);
        Integer index = QuestMobHelper.getIndex(entity);
        if (index == null) {
            return false;
        }
        return questPlayer.getCurrentQuests().containsKey(quest) && questPlayer.getCurrentStage(quest) == stage &&
                questPlayer.getQuestDataOrDefault(quest).getMobNumKilled().get(index) < stage.getMobNumToKill().get(index);
    }

    /**
     * Gets the quest with the given id
     *
     * @param questId <p>The id of the quest to get</p>
     * @return <p>The quest, or null if no such quest exists</p>
     */
    private @Nullable Quest getQuest(@NotNull String questId) {
        for (Quest quest : questsAPI.getLoadedQuests()) {
            if (quest.getId().equalsIgnoreCase(questId)) {
                return quest;
            }
        }
        return null;
    }

    /**
     * Cancels a cancellable event if the given entity has been spawned by this plugin
     *
     * @param cancellable <p>A cancellable event</p>
     * @param entity      <p>The entity to check</p>
     */
    private void cancelIfQuestMob(@NotNull Cancellable cancellable, @NotNull Entity entity) {
        if (QuestMobHelper.isSpawnedMob(entity)) {
            cancellable.setCancelled(true);
        }
    }

}
